import datetime
import fnmatch
import logging
import os
import sys
import requests
import urllib3
import cgi

from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, TypedDict, Union, cast
from urllib import parse as urlparse

from requests.adapters import HTTPAdapter, Retry
from requests_toolbelt.multipart.encoder import MultipartEncoderMonitor

from .interfaces import QfcException, QfcRequest, QfcRequestException
from .utils import calc_etag, log, add_trailing_slash_to_url
from pathvalidate import is_valid_filepath


logger = logging.getLogger(__file__)

if sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata


try:
    __version__ = metadata.version("qfieldcloud_sdk")
except metadata.PackageNotFoundError:
    __version__ = "dev"


DEFAULT_PAGINATION_LIMIT = 20
"""int: Defines the default limit for pagination, set to `20`."""


class FileTransferStatus(str, Enum):
    """Represents the status of a file transfer.

    Attributes:
        PENDING: The transfer is pending.
        SUCCESS: The transfer was successful.
        FAILED: The transfer failed.
    """

    PENDING = "PENDING"
    SUCCESS = "SUCCESS"
    FAILED = "FAILED"


class FileTransferType(str, Enum):
    """Represents the type of file transfer.

    The PACKAGE transfer type is used only internally in QFieldCloud workers, so it should never be used by other API clients.

    Attributes:
        PROJECT: Refers to a project file.
        PACKAGE: Refers to a package Type.
    """

    PROJECT = "project"
    PACKAGE = "package"


class JobTypes(str, Enum):
    """Represents the types of jobs that can be processed on QFieldCloud.

    Attributes:
        PACKAGE: Refers to a packaging job.
        APPLY_DELTAS: Refers to applying deltas (differences).
        PROCESS_PROJECTFILE: Refers to processing a project file.
    """

    PACKAGE = "package"
    APPLY_DELTAS = "delta_apply"
    PROCESS_PROJECTFILE = "process_projectfile"
    CREATE_PROJECT = "create_project"


class ProjectCollaboratorRole(str, Enum):
    """Defines roles for project collaborators.

    See project collaborator roles documentation: https://docs.qfield.org/reference/qfieldcloud/permissions/#roles_1

    Attributes:
        ADMIN: Administrator role.
        MANAGER: Manager role.
        EDITOR: Editor role.
        REPORTER: Reporter role.
        READER: Reader role.
    """

    ADMIN = "admin"
    MANAGER = "manager"
    EDITOR = "editor"
    REPORTER = "reporter"
    READER = "reader"


class OrganizationMemberRole(str, Enum):
    """Defines roles for organization members.

    See organization member roles documentation: https://docs.qfield.org/reference/qfieldcloud/permissions/#roles_2

    Attributes:
        ADMIN: Administrator role.
        MEMBER: Member role.
    """

    ADMIN = "admin"
    MEMBER = "member"


class CollaboratorModel(TypedDict):
    """Represents the structure of a project collaborator in the QFieldCloud system.

    Attributes:
        collaborator: The collaborator's identifier.
        role: The role of the collaborator.
        project_id: The associated project identifier.
        created_by: The user who created the collaborator entry.
        updated_by: The user who last updated the collaborator entry.
        created_at: The timestamp when the collaborator entry was created.
        updated_at: The timestamp when the collaborator entry was last updated.
    """

    collaborator: str
    role: ProjectCollaboratorRole
    project_id: str
    created_by: str
    updated_by: str
    created_at: datetime.datetime
    updated_at: datetime.datetime


class OrganizationMemberModel(TypedDict):
    """Represents the structure of an organization member in the QFieldCloud system.

    Attributes:
        member: The member's identifier.
        role: The role of the member.
        organization: The associated organization identifier.
        is_public: A boolean indicating if the membership is public.
    """

    member: str
    role: OrganizationMemberRole
    organization: str
    is_public: bool
    # TODO future work that can be surely expected, check QF-4535
    # created_by: str
    # updated_by: str
    # created_at: datetime.datetime
    # updated_at: datetime.datetime


class TeamModel(TypedDict):
    """Represents the structure of a team within an organization in the QFieldCloud system.

    Attributes:
        team: The team's identifier.
        organization: The associated organization identifier.
        members: A list of member identifiers belonging to the team.
    """

    team: str
    organization: str
    members: List[str]


class TeamMemberModel(TypedDict):
    """Represents the structure of a team member in the QFieldCloud system.

    Attributes:
        member: The member's identifier.
    """

    member: str


class Pagination:
    """The Pagination class allows for controlling and managing pagination of results within the QFieldCloud SDK.

    Args:
        limit: The maximum number of items to return. Defaults to None.
        offset: The starting point from which to return items. Defaults to None.

    Attributes:
        limit: The maximum number of items to return.
        offset: The starting point from which to return items.
    """

    limit: Optional[int] = None
    offset: Optional[int] = None

    def __init__(
        self, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> None:
        self.limit = limit
        self.offset = offset

    @property
    def is_empty(self) -> bool:
        """Whether both limit and offset are None, indicating no pagination settings.

        Returns:
            True if both limit and offset are None, False otherwise.
        """
        return self.limit is None and self.offset is None


class DeltaPushResponse(TypedDict):
    """Represents the structure of the response for pushing a delta file.

    Attributes:
        status: The status of the response.
        message: A message providing additional information about the response.
        details: Additional details about the delta push operation.
    """

    status: str
    message: Optional[str]
    details: Optional[Dict[str, Any]]


class Client:
    """The core component of the QFieldCloud SDK, providing methods for interacting with the QFieldCloud platform.

    The client provides methods for authentication, project management, file management, and more.

    It is configured with retries for GET requests on specific 5xx HTTP status codes.

    Args:
        url: The base URL for the QFieldCloud API. Defaults to `QFIELDCLOUD_URL` environment variable if empty.
        verify_ssl: Whether to verify SSL certificates. Defaults to True.
        token: The authentication token for API access. Defaults to `QFIELDCLOUD_TOKEN` environment variable if empty.

    Raises:
        QfcException: If the `url` is not provided either directly or through the environment variable.

    Attributes:
        session: The session object to maintain connections.
        url: The base URL for the QFieldCloud API.
        token: The authentication token for API access.
        verify_ssl: Whether to verify SSL certificates.
    """

    session: requests.Session

    url: str

    token: str

    verify_ssl: bool

    def __init__(
        self,
        url: str = "",
        verify_ssl: bool = True,
        token: str = "",
    ) -> None:
        self.session = requests.Session()
        # retries should be only on GET and only if error 5xx
        retries = Retry(
            total=5,
            backoff_factor=0.1,
            allowed_methods=["GET"],
            # skip 501, as it is "Not Implemented", no point to retry
            status_forcelist=[500, 502, 503, 504],
        )
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

        self.url = url or os.environ.get("QFIELDCLOUD_URL", "")
        self.token = token or os.environ.get("QFIELDCLOUD_TOKEN", "")
        self.verify_ssl = verify_ssl

        if not self.verify_ssl:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        if not self.url:
            raise QfcException(
                "Cannot create a new QFieldCloud client without a url passed in the constructor or as environment variable QFIELDCLOUD_URL"
            )

    def login(self, username: str, password: str) -> Dict[str, Any]:
        """Logs in with the provided username and password.

        Args:
            username: The username or email used to register.
            password: The password associated with the username.

        Returns:
            Authentication token and additional metadata.

        Example:
            ```python
            client = sdk.Client(url="https://app.qfield.cloud/api/v1/")
            client.login("ninjamaster", "secret_password123")
            ```
        """
        resp = self._request(
            "POST",
            "auth/login",
            data={
                "username": username,
                "password": password,
            },
            skip_token=True,
        )

        payload = resp.json()

        self.token = payload["token"]

        return payload

    def logout(self) -> None:
        """Logs out from the current session, invalidating the authentication token.

        Example:
            ```python
            client.logout()
            ```
        """
        resp = self._request("POST", "auth/logout")

        return resp.json()

    def check_server_status(self) -> Dict[str, Any]:
        """Checks the status of the QFieldCloud server.

        This endpoint usually provides information of the server health.

        Returns:
            A dictionary containing the server status information (e.g., {"storage": "ok", ...}).

        Example:
            ```python
            client = sdk.Client(url="[https://app.qfield.cloud/api/v1/](https://app.qfield.cloud/api/v1/)")
            status = client.check_server_status()
            print(f"Server status: {status}")
            ```
        """
        resp = self._request(
            "GET",
            "status/",
            skip_token=True,
        )
        return resp.json()

    def list_projects(
        self,
        include_public: bool = False,
        pagination: Pagination = Pagination(),
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """List projects accessible by the current user. Optionally include all public projects.

        Args:
            include_public: Whether to include public projects in the list. Defaults to False.
            pagination: Pagination settings for the request. Defaults to an empty Pagination instance.

        Returns:
            A list of dictionaries containing project details.

        Example:
            ```python
            client.list_projects()
            ```
        """
        params = {
            "include-public": str(int(include_public)),  # type: ignore
        }

        payload = self._request_json(
            "GET", "projects", params=params, pagination=pagination
        )
        return cast(List, payload)

    def get_project(
        self,
        project_id: str,
    ) -> Dict[str, Any]:
        """Get project data.

        Args:
            project_id: the project data to get data for.

        Returns:
            A dictionary containing project details.

        Example:
            ```python
            client.get_project(project_id)
            ```
        """
        payload = self._request_json("GET", f"projects/{project_id}")

        return cast(Dict, payload)

    def get_project_seed(
        self,
        project_id: str,
    ) -> Dict[str, Any]:
        """Get project seed data.

        Args:
            project_id: the project data to get seed data for.

        Returns:
            A dictionary containing project seed.

        Example:
            ```python
            client.get_project_seed(project_id)
            ```
        """
        payload = self._request_json("GET", f"projects/{project_id}/seed")

        return cast(Dict, payload)

    def get_project_seed_xlsform(
        self,
        project_id: str,
        destination_dir: str,
    ) -> str | None:
        """Get project seed XLSForm file content.

        Args:
            project_id: the project data to get seed XLSForm for.

        Returns:
            The name of the downloaded XLSForm file.

        Example:
            ```python
            client.get_project_seed_xlsform(project_id)
            ```
        """

        resp = self._request("GET", f"projects/{project_id}/seed/xlsform")

        if resp.status_code != 200:
            return None

        content_disposition = resp.headers.get("Content-Disposition", "")

        if not content_disposition:
            logger.warning(
                "Response has no `Content-Disposition` header. Skip download of XLSForm file!"
            )

            return None

        _value, params = cgi.parse_header(content_disposition)
        filename = params.get("filename")

        if not filename:
            logger.warning(
                "Response has no filename in `Content-Disposition` header. Skip download of XLSForm file!"
            )

            return None

        path = Path(destination_dir).joinpath(filename)
        path.write_bytes(resp.content)

        return str(path)

    def list_remote_files(
        self, project_id: str, skip_metadata: bool = True
    ) -> List[Dict[str, Any]]:
        """List project files.

        Args:
            project_id: Project ID.
            skip_metadata: Whether to skip fetching metadata for the files. Defaults to True.

        Returns:
            A list of file details.

        Example:
            ```python
            client.list_remote_files("123e4567-e89b-12d3-a456-426614174000", False)
            ```
        """
        params = {}

        if skip_metadata:
            params["skip_metadata"] = "1"

        resp = self._request("GET", f"files/{project_id}", params=params)
        remote_files = resp.json()
        # TODO remove this temporary decoration with `etag` key
        remote_files = list(map(lambda f: {"etag": f["md5sum"], **f}, remote_files))

        return remote_files

    def create_project(
        self,
        name: str,
        owner: Optional[str] = None,
        description: str = "",
        is_public: bool = False,
    ) -> Dict[str, Any]:
        """Create a new project.

        Args:
            name: The name of the new project.
            owner: The owner of the project. When None, the project will be owned by the currently logged-in user. Defaults to None.
            description: A description of the project. Defaults to an empty string.
            is_public: Whether the project should be public. Defaults to False.

        Returns:
            A dictionary containing the details of the created project.

        Example:
            ```python
            client.create_project(
                "Tree_Survey", "My_Organization_Clan", "Description"
            )
            ```
        """
        resp = self._request(
            "POST",
            "projects",
            data={
                "name": name,
                "owner": owner,
                "description": description,
                "is_public": int(is_public),
            },
        )

        return resp.json()

    def delete_project(self, project_id: str) -> requests.Response:
        """Delete a project.

        Args:
            project_id: Project ID.

        Returns:
            The response object from the file delete request.

        Example:
            ```python
            client.delete_project("123e4567-e89b-12d3-a456-426614174000")
            ```
        """
        resp = self._request("DELETE", f"projects/{project_id}")

        return resp

    def patch_project(
        self,
        project_id: str,
        name: Optional[str] = None,
        owner: Optional[str] = None,
        description: Optional[str] = None,
        is_public: Optional[bool] = None,
    ) -> Dict[str, Any]:
        """Update a project.

        Args:
            project_id (str): Project ID.
            name (str | None): if passed, the new name. Defaults to None.
            owner (str | None, optional): if passed, the new owner. Defaults to None.
            description (str, optional): if passed, the new description. Defaults to None.
            is_public (bool, optional): if passed, the new public setting. Defaults to None.

        Returns:
            Dict[str, Any]: the updated project

        Example:
            ```python
            client.patch_project(description="New Description")
            ```
        """
        project_data: dict[str, Any] = {}

        if name:
            project_data["name"] = name

        if description:
            project_data["description"] = description

        if owner:
            project_data["owner"] = owner

        if is_public:
            project_data["is_public"] = is_public

        resp = self._request(
            "PATCH",
            f"projects/{project_id}",
            project_data,
        )

        return resp.json()

    def upload_files(
        self,
        project_id: str,
        upload_type: FileTransferType,
        project_path: str,
        filter_glob: str,
        throw_on_error: bool = False,
        show_progress: bool = False,
        force: bool = False,
        job_id: str = "",
    ) -> List[Dict]:
        """Upload files to a QFieldCloud project.

        Args:
            project_id: Project ID.
            upload_type: The type of file transfer (PROJECT or PACKAGE).
            project_path: The local directory containing the files to upload.
            filter_glob: A glob pattern to filter which files to upload.
            throw_on_error: Whether to raise an error if a file fails to upload. Defaults to False.
            show_progress: Whether to display a progress bar during upload. Defaults to False.
            force: Whether to force upload all files, even if they exist remotely. Defaults to False.
            job_id: The job ID, required if `upload_type` is PACKAGE. Defaults to an empty string.

        Returns:
            A list of dictionaries with information about the uploaded files.

        Example:
            ```python
            client.upload_files(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                upload_type=sdk.FileTransferType.PROJECT,
                project_path="/home/ninjamaster/QField/cloud/Tree_Survey",
                filter_glob="*",
                throw_on_error=True,
                show_progress=True,
                force=True
            )
            ```
        """
        if not filter_glob:
            filter_glob = "*"

        local_files = self.list_local_files(project_path, filter_glob)

        # we should always upload all package files
        if upload_type == FileTransferType.PACKAGE:
            force = True

        files_to_upload = []
        if force:
            files_to_upload = local_files
        else:
            remote_files = self.list_remote_files(project_id)

            if len(remote_files) == 0:
                files_to_upload = local_files
            else:
                for local_file in local_files:
                    remote_file = None
                    for f in remote_files:
                        if f["name"] == local_file["name"]:
                            remote_file = f
                            break

                    etag = calc_etag(local_file["absolute_filename"])
                    if remote_file and remote_file.get("etag", None) == etag:
                        continue

                    files_to_upload.append(local_file)

        if not files_to_upload:
            return files_to_upload

        for file in files_to_upload:
            try:
                local_filename = Path(file["absolute_filename"])
                self.upload_file(
                    project_id,
                    upload_type,
                    local_filename,
                    file["name"],
                    show_progress,
                    job_id,
                )
                file["status"] = FileTransferStatus.SUCCESS
            except Exception as err:
                file["status"] = FileTransferStatus.FAILED
                file["error"] = err

                if throw_on_error:
                    raise err
                else:
                    continue

        return local_files

    def upload_file(
        self,
        project_id: str,
        upload_type: FileTransferType,
        local_filename: Path,
        remote_filename: Path,
        show_progress: bool,
        job_id: str = "",
    ) -> requests.Response:
        """Upload a single file to a project.

        Args:
            project_id: Project ID.
            upload_type: The type of file transfer.
            local_filename: The path to the local file to upload.
            remote_filename: The path where the file should be stored remotely.
            show_progress: Whether to display a progress bar during upload.
            job_id: The job ID, required if `upload_type` is PACKAGE. Defaults to an empty string.

        Raises:
            pathvalidate.ValidationError: Raised when the uploaded file does not have a valid filename.

        Returns:
            The response object from the upload request.

        Example:
            ```python
            client.upload_file(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                upload_type=FileTransferType.PROJECT,
                local_filename="/home/ninjamaster/QField/cloud/Tree_Survey/trees.gpkg",
                remote_filename="trees.gpkg",
                show_progress=True
            )
            ```
        """
        # if the filepath is invalid, it will throw a new error `pathvalidate.ValidationError`
        is_valid_filepath(str(local_filename))

        local_file_size = local_filename.stat().st_size
        with open(local_filename, "rb") as local_file:
            encoder_params = {}

            if show_progress:
                from tqdm import tqdm

                progress_bar = tqdm(
                    total=local_file_size,
                    unit_scale=True,
                    unit="B",
                    desc=f'Uploading "{remote_filename}"...',
                )

                def cb(monitor: MultipartEncoderMonitor) -> None:
                    progress_bar.n = monitor.bytes_read
                    progress_bar.refresh()

                encoder_params["callback"] = cb
            else:
                logger.info(f'Uploading file "{remote_filename}"…')

            multipart_data = MultipartEncoderMonitor.from_fields(
                fields={
                    "file": (
                        str(remote_filename),
                        local_file,
                        None,
                    ),
                },
                **encoder_params,
            )

            if upload_type == FileTransferType.PROJECT:
                url = f"files/{project_id}/{remote_filename}"
            elif upload_type == FileTransferType.PACKAGE:
                if not job_id:
                    raise QfcException(
                        'When the upload type is "package", you must pass the "job_id" parameter.'
                    )

                url = f"packages/{project_id}/{job_id}/files/{remote_filename}"
            else:
                raise NotImplementedError()

            return self._request(
                "POST",
                url,
                data=multipart_data,
                headers={
                    "Content-Type": multipart_data.content_type,
                    "Accept": "application/json",
                },
            )

    def download_project(
        self,
        project_id: str,
        local_dir: str,
        filter_glob: str = "*",
        throw_on_error: bool = False,
        show_progress: bool = False,
        force_download: bool = False,
    ) -> List[Dict]:
        """Download the specified project files into the destination dir.

        Args:
            project_id: Project ID.
            local_dir: destination directory where the files will be downloaded
            filter_glob: if specified, download only project files which match the glob, otherwise download all
            throw_on_error: Throw if download error occurres. Defaults to False.
            show_progress: Show progress bar in the console. Defaults to False.
            force_download: Download file even if it already exists locally. Defaults to False.

        Returns:
            A list of dictionaries with information about the downloaded files.

        Example:
            ```python
            client.download_project(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                local_dir="/home/ninjamaster/QField/cloud/Tree_Survey",
                filter_glob="*",
                show_progress=True,
                force_download=True
            )
            ```
        """
        files = self.list_remote_files(project_id)

        return self.download_files(
            files,
            project_id,
            FileTransferType.PROJECT,
            local_dir,
            filter_glob,
            throw_on_error,
            show_progress,
            force_download,
        )

    def list_jobs(
        self,
        project_id: str,
        job_type: Optional[JobTypes] = None,
        pagination: Pagination = Pagination(),
    ) -> List[Dict[str, Any]]:
        """List project jobs.

        Args:
            project_id: Project ID.
            job_type: The type of job to filter by. If set to None, list all jobs. Defaults to None.
            pagination: Pagination settings. Defaults to a new Pagination object.

        Returns:
            A list of dictionaries representing the jobs.

        Example:
            ```python
                client.list_jobs(
                    project_id="123e4567-e89b-12d3-a456-426614174000",
                    job_type=JobTypes.PACKAGE
                )
            ```
        """
        payload = self._request_json(
            "GET",
            "jobs/",
            {
                "project_id": project_id,
                "type": job_type.value if job_type else None,
            },
            pagination=pagination,
        )
        return cast(List, payload)

    def job_trigger(
        self, project_id: str, job_type: JobTypes, force: bool = False
    ) -> Dict[str, Any]:
        """Trigger a new job for given project.

        Args:
            project_id: Project ID.
            job_type (JobTypes): The type of job to trigger.
            force: Whether to force the job execution. Defaults to False.

        Returns:
            A dictionary containing the job information.

        Example:
            ```python
            client.job_trigger(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                job_type=JobTypes.PACKAGE,
                force=True
            )
            ```
        """
        resp = self._request(
            "POST",
            "jobs/",
            {
                "project_id": project_id,
                "type": job_type.value,
                "force": int(force),
            },
        )

        return resp.json()

    def job_status(self, job_id: str) -> Dict[str, Any]:
        """Get the status of a job.

        Args:
            job_id: The ID of the job.

        Returns:
            A dictionary containing the job status.

        Example:
            ```python
            client.job_status("123e4567-e89b-12d3-a456-426614174000")
            ```
        """
        resp = self._request("GET", f"jobs/{job_id}")

        return resp.json()

    def push_delta(self, project_id: str, delta_filename: str) -> DeltaPushResponse:
        """Push a delta file to a project.

        Args:
            project_id: Project ID.
            delta_filename: Path to the delta JSON file.

        Returns:
            A DeltaPushResponse containing the response from the server.

        Example:
            ```python
            client.push_delta(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                delta_filename="/home/ninjamaster/QField/cloud/Tree_Survey/deltas.json"
            )
            ```
        """
        with open(delta_filename, "r") as delta_file:
            files = {"file": delta_file}
            response = self._request(
                "POST",
                f"deltas/{project_id}/",
                files=files,
            )

        return cast(DeltaPushResponse, response)

    def delete_files(
        self,
        project_id: str,
        glob_patterns: List[str],
        throw_on_error: bool = False,
        finished_cb: Optional[Callable] = None,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """Delete project files.

        Args:
            project_id: Project ID.
            glob_patterns (list[str]): Delete only files matching one the glob patterns.
            throw_on_error: Throw if delete error occurres. Defaults to False.
            finished_cb (Callable, optional): Deprecated. Defaults to None.

        Raises:
            QFieldCloudException: if throw_on_error is True, throw an error if a download request fails.

        Returns:
            Deleted files by glob pattern.

        Example:
            ```python
            client.delete_files(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                glob_patterns=["*.csv", "*.jpg"],
                throw_on_error=True
            )
            ```
        """
        project_files = self.list_remote_files(project_id)
        glob_results: Dict[str, List[Dict[str, Any]]] = {}

        log(f"Project '{project_id}' has {len(project_files)} file(s).")

        for glob_pattern in glob_patterns:
            glob_results[glob_pattern] = []

            for file in project_files:
                if not fnmatch.fnmatch(file["name"], glob_pattern):
                    continue

                if "status" in file:
                    # file has already been matched by a previous glob pattern
                    continue

                file["status"] = FileTransferStatus.PENDING
                glob_results[glob_pattern].append(file)

        for glob_pattern, files in glob_results.items():
            if not files:
                log(f"Glob pattern '{glob_pattern}' did not match any files.")
                continue

            for file in files:
                try:
                    resp = self._request(
                        "DELETE",
                        f"files/{project_id}/{file['name']}",
                        stream=True,
                    )
                    file["status"] = FileTransferStatus.SUCCESS
                except QfcRequestException as err:
                    resp = err.response

                    logger.info(
                        f"{resp.request.method} {resp.url} got HTTP {resp.status_code}"
                    )

                    file["status"] = FileTransferStatus.FAILED
                    file["error"] = err

                    log(f'File "{file["name"]}" failed to delete:\n{file["error"]}')

                    if throw_on_error:
                        continue
                    else:
                        raise err
                finally:
                    if callable(finished_cb):
                        finished_cb(file)

        files_deleted = 0
        files_failed = 0
        for files in glob_results.values():
            for file in files:
                log(f"{file['status']}\t{file['name']}")

                if file["status"] == FileTransferStatus.SUCCESS:
                    files_deleted += 1
                elif file["status"] == FileTransferStatus.SUCCESS:
                    files_failed += 1

        log(f"{files_deleted} file(s) deleted, {files_failed} file(s) failed to delete")

        return glob_results

    def delete_file(
        self,
        project_id: str,
        filename: str,
        version_id: Optional[str],
    ) -> requests.Response:
        headers = {}

        if version_id:
            headers["x-file-version"] = version_id

        resp = self._request(
            "DELETE",
            f"files/{project_id}/{filename}",
            headers=headers,
        )

        return resp

    def package_latest(self, project_id: str) -> Dict[str, Any]:
        """Check the latest packaging status of a project.

        Args:
            project_id: Project ID.

        Returns:
            A dictionary containing the latest packaging status.

        Example:
            ```python
            client.package_latest("123e4567-e89b-12d3-a456-426614174000")
            ```
        """
        resp = self._request("GET", f"packages/{project_id}/latest/")

        return resp.json()

    def package_download(
        self,
        project_id: str,
        local_dir: str,
        filter_glob: str = "*",
        throw_on_error: bool = False,
        show_progress: bool = False,
        force_download: bool = False,
    ) -> List[Dict]:
        """Download the specified project packaged files into the destination dir.

        Args:
            project_id: Project ID.
            local_dir: destination directory where the files will be downloaded
            filter_glob: if specified, download only packaged files which match the glob, otherwise download all
            throw_on_error: Throw if download error occurres. Defaults to False.
            show_progress: Show progress bar in the console. Defaults to False.
            force_download: Download file even if it already exists locally. Defaults to False.

        Returns:
            A list of dictionaries with information about the downloaded files.

        Example:
            ```python
            client.package_download(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                local_dir="/home/ninjamaster/QField/cloud/Tree_Survey",
                filter_glob="*",
                show_progress=True
            )
            ```
        """
        project_status = self.package_latest(project_id)

        if project_status["status"] != "finished":
            raise QfcException(
                "The project has not been successfully packaged yet. Please use `qfieldcloud-cli package-trigger {project_id}` first."
            )

        resp = self._request("GET", f"packages/{project_id}/latest/")

        return self.download_files(
            resp.json()["files"],
            project_id,
            FileTransferType.PACKAGE,
            local_dir,
            filter_glob,
            throw_on_error,
            show_progress,
            force_download,
        )

    def download_files(
        self,
        files: List[Dict],
        project_id: str,
        download_type: FileTransferType,
        local_dir: str,
        filter_glob: str = "*",
        throw_on_error: bool = False,
        show_progress: bool = False,
        force_download: bool = False,
    ) -> List[Dict]:
        """Download project files.

        Args:
            files : A list of file dicts, specifying which files to download.
            project_id: Project ID.
            download_type: File transfer type which specifies what should be the download url.
            local_dir: Local destination directory
            filter_glob: Download only files matching the glob pattern. If None download all. Defaults to None.
            throw_on_error: Throw if download error occurres. Defaults to False.
            show_progress: Show progress bar in the console. Defaults to False.
            force_download: Download file even if it already exists locally. Defaults to False.

        Raises:
            QFieldCloudException: if throw_on_error is True, throw an error if a download request fails.

        Returns:
            A list of file dicts.

        Example:
            ```python
            client.download_files(
                files=[{"name": "trees.gpkg"}, {"name": "roads.gpkg"],
                project_id="123e4567-e89b-12d3-a456-426614174000",
                download_type=FileTransferType.PROJECT,
                local_dir="/home/ninjamaster/QField/cloud/Tree_Survey",
                filter_glob="*",
                show_progress=True
            )
            ```
        """
        if not filter_glob:
            filter_glob = "*"

        files_to_download: List[Dict[str, Any]] = []

        for file in files:
            if fnmatch.fnmatch(file["name"], filter_glob):
                file["status"] = FileTransferStatus.PENDING
                files_to_download.append(file)

        for file in files_to_download:
            local_filename = Path(f"{local_dir}/{file['name']}")
            etag = None
            if not force_download:
                etag = file.get("etag", None)

            try:
                self.download_file(
                    project_id,
                    download_type,
                    local_filename,
                    file["name"],
                    show_progress,
                    etag,
                )
                file["status"] = FileTransferStatus.SUCCESS
            except QfcRequestException as err:
                resp = err.response

                logger.info(
                    f"{resp.request.method} {resp.url} got HTTP {resp.status_code}"
                )

                file["status"] = FileTransferStatus.FAILED
                file["error"] = err

                if throw_on_error:
                    raise err
                else:
                    continue

        return files_to_download

    def download_file(
        self,
        project_id: str,
        download_type: FileTransferType,
        local_filename: Path,
        remote_filename: Path,
        show_progress: bool,
        remote_etag: Optional[str] = None,
    ) -> Optional[requests.Response]:
        """Download a single project file.

        Args:
            project_id: Project ID.
            download_type: File transfer type which specifies what should be the download URL
            local_filename: Local filename
            remote_filename: Remote filename
            show_progress: Show progressbar in the console
            remote_etag: The ETag of the remote file. If is None, the download of the file happens even if it already exists locally. Defaults to `None`.

        Raises:
            NotImplementedError: Raised if unknown `download_type` is passed

        Returns:
            The response object.

        Example:
            ```python
            client.download_file(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                download_type=FileTransferType.PROJECT,
                local_filename="/home/ninjamaster/QField/cloud/Tree_Survey/trees.gpkg",
                remote_filename="trees.gpkg",
                show_progress=True
            )
            ```
        """

        headers: dict[str, str] = {}
        if remote_etag and local_filename.exists():
            local_etag = calc_etag(str(local_filename))

            if local_etag == remote_etag:
                if show_progress:
                    print(
                        f"{remote_filename}: Already present locally. Download skipped."
                    )
                else:
                    logger.info(
                        f'Skipping download of "{remote_filename}" because it is already present locally'
                    )
                return None

            headers["If-None-Match"] = local_etag

        if download_type == FileTransferType.PROJECT:
            url = f"files/{project_id}/{remote_filename}"
        elif download_type == FileTransferType.PACKAGE:
            url = f"packages/{project_id}/latest/files/{remote_filename}"
        else:
            raise NotImplementedError()

        resp = self._request("GET", url, stream=True, headers=headers)

        # Since we are sending the `If-None-Match` header to check the `ETag` of the remote file,
        # we shall expect HTTP 304 in case the local and remote `ETag` match.
        # The early return will prevent overwriting the file with empty contents of the 304 response.
        if resp.status_code == 304:
            return None

        if not local_filename.parent.exists():
            local_filename.parent.mkdir(parents=True)

        with open(local_filename, "wb") as f:
            download_file = f
            if show_progress:
                from tqdm import tqdm
                from tqdm.utils import CallbackIOWrapper

                content_length = int(resp.headers.get("content-length", 0))
                progress_bar = tqdm(
                    total=content_length,
                    unit_scale=True,
                    desc=str(remote_filename),
                )
                download_file = CallbackIOWrapper(progress_bar.update, f, "write")
            else:
                logger.info(f'Downloading file "{remote_filename}"…')

            for chunk in resp.iter_content(chunk_size=8192):
                # filter out keep-alive new chunks
                if chunk:
                    download_file.write(chunk)

        return resp

    def list_local_files(
        self, root_path: str, filter_glob: str
    ) -> List[Dict[str, Any]]:
        """
        Returns a list of dicts with information about local files. Usually used before uploading files.
        NOTE: files and dirs starting with leading dot (.) or ending in tilde (~) will be ignored.

        Example:
            ```python
            client.list_local_files("/home/ninjamaster/QField/cloud/Tree_Survey", "*.gpkg")
            ```
        """
        if not filter_glob:
            filter_glob = "*"

        files: List[Dict[str, Any]] = []
        for path in Path(root_path).rglob(filter_glob):
            if not path.is_file():
                continue

            basename = path.relative_to(root_path).name
            if basename.startswith(".") or basename.endswith("~"):
                continue

            relative_name = path.relative_to(root_path)
            files.append(
                {
                    "name": str(relative_name),
                    "absolute_filename": str(path),
                    "status": FileTransferStatus.PENDING,
                    "error": None,
                }
            )

        # upload the QGIS project file at the end
        files.sort(key=lambda f: Path(f["name"]).suffix.lower() in (".qgs", ".qgz"))

        return files

    def get_project_collaborators(self, project_id: str) -> List[CollaboratorModel]:
        """Gets a list of project collaborators.

        Args:
            project_id: Project ID.

        Returns:
            The list of project collaborators.

        Example:
            ```python
            client.get_project_collaborators("123e4567-e89b-12d3-a456-426614174000")
            ```
        """
        collaborators = cast(
            List[CollaboratorModel],
            self._request_json("GET", f"/collaborators/{project_id}"),
        )

        return collaborators

    def add_project_collaborator(
        self, project_id: str, username: str, role: ProjectCollaboratorRole
    ) -> CollaboratorModel:
        """Adds a project collaborator.

        Args:
            project_id: Project ID.
            username: username of the collaborator to be added
            role: the role of the collaborator. One of: `reader`, `reporter`, `editor`, `manager` or `admin`

        Returns:
            The added project collaborator.

        Example:
            ```python
            client.add_project_collaborator(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                username="ninja_001",
                role=ProjectCollaboratorRole.EDITOR
            )
            ```
        """
        collaborator = cast(
            CollaboratorModel,
            self._request_json(
                "POST",
                f"/collaborators/{project_id}",
                {
                    "collaborator": username,
                    "role": role,
                },
            ),
        )

        return collaborator

    def remove_project_collaborator(self, project_id: str, username: str) -> None:
        """Removes a project collaborator.

        Args:
            project_id: Project ID.
            username: the username of the collaborator to be removed

        Example:
            ```python
            client.remove_project_collaborator(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                username="ninja_007"
            )
            ```
        """
        self._request("DELETE", f"/collaborators/{project_id}/{username}")

    def patch_project_collaborators(
        self, project_id: str, username: str, role: ProjectCollaboratorRole
    ) -> CollaboratorModel:
        """Change an already existing project collaborator.

        Args:
            project_id: Project ID.
            username: the username of the collaborator to be patched
            role: the new role of the collaborator

        Returns:
            The updated project collaborator.

        Example:
            ```python
            client.patch_project_collaborators(
                project_id="123e4567-e89b-12d3-a456-426614174000",
                username="ninja_001",
                role=ProjectCollaboratorRole.MANAGER
            )
            ```
        """
        collaborator = cast(
            CollaboratorModel,
            self._request_json(
                "PATCH",
                f"/collaborators/{project_id}/{username}",
                {
                    "role": role,
                },
            ),
        )

        return collaborator

    def get_organization_members(
        self, organization: str
    ) -> List[OrganizationMemberModel]:
        """Gets a list of organization members.

        Args:
            organization: organization username

        Returns:
            The list of organization members.

        Example:
            ```python
            client.get_organization_members("My_Organization_Clan")
            ```
        """
        members = cast(
            List[OrganizationMemberModel],
            self._request_json("GET", f"/members/{organization}"),
        )

        return members

    def add_organization_member(
        self,
        organization: str,
        username: str,
        role: OrganizationMemberRole,
        is_public: bool,
    ) -> OrganizationMemberModel:
        """Adds an organization member.

        Args:
            organization: organization username
            username: username of the member to be added
            role: the role of the member. One of: `admin` or `member`.
            is_public: whether the membership should be public.

        Returns:
            The added organization member.

        Example:
            ```python
            client.add_organization_member(
                organization="My_Organization_Clan",
                username="ninja_001",
                role=OrganizationMemberRole.MEMBER,
                is_public=False
            )
            ```
        """
        member = cast(
            OrganizationMemberModel,
            self._request_json(
                "POST",
                f"/members/{organization}",
                {
                    "member": username,
                    "role": role,
                    "is_public": is_public,
                },
            ),
        )

        return member

    def remove_organization_members(self, project_id: str, username: str) -> None:
        """Removes an organization member.

        Args:
            project_id: Project ID.
            username: the username of the member to be removed

        Example:
            ```python
            client.remove_organization_members(
                organization="My_Organization_Clan",
                username="ninja_007"
            )
            ```
        """
        self._request("DELETE", f"/members/{project_id}/{username}")

    def patch_organization_members(
        self, project_id: str, username: str, role: OrganizationMemberRole
    ) -> OrganizationMemberModel:
        """Change an already existing organization member.

        Args:
            project_id: Project ID.
            username: the username of the member to be patched
            role: the new role of the member

        Returns:
            The updated organization member.

        Example:
            ```python
            client.patch_organization_members(
                organization="My_Organization_Clan",
                username="ninja_001",
                role=OrganizationMemberRole.ADMIN
            )
            ```
        """
        member = cast(
            OrganizationMemberModel,
            self._request_json(
                "PATCH",
                f"/members/{project_id}/{username}",
                {
                    "role": role,
                },
            ),
        )

        return member

    def get_organization_teams(self, organization: str) -> List[TeamModel]:
        """Gets a list of organization teams.

        Args:
            organization: The name of the organization.

        Returns:
            A list of TeamModel objects representing the organization's teams.

        Example:
            ```python
            teams = client.get_organization_teams(organization="My_Organization_Clan")
            for team in teams:
                print(team.team)
            ```
        """
        teams = cast(
            List[TeamModel],
            self._request_json("GET", f"organizations/{organization}/teams/"),
        )
        return teams

    def create_organization_team(self, organization: str, team_name: str) -> TeamModel:
        """Creates a new organization team.

        Args:
            organization: The name of the organization.
            team_name: The name of the new team.

        Returns:
            A TeamModel object representing the newly created team.

        Example:
            ```python
            client.create_organization_team(organization="My_Organization_Clan", team_name="surveyors")
            ```
        """
        team = cast(
            TeamModel,
            self._request_json(
                "POST",
                f"organizations/{organization}/teams/",
                {"team": team_name},
            ),
        )
        return team

    def get_organization_team(self, organization: str, team_name: str) -> TeamModel:
        """Gets an organization team by name.

        Args:
            organization: The name of the organization.
            team_name: The name of the team to retrieve.

        Returns:
            A TeamModel object representing the requested team.

        Example:
            ```python
            client.get_organization_team(organization="My_Organization_Clan", team_name="surveyors")
            ```
        """
        team = cast(
            TeamModel,
            self._request_json(
                "GET", f"organizations/{organization}/teams/{team_name}/"
            ),
        )
        return team

    def patch_organization_team(
        self, organization: str, team_name: str, new_team_name: str
    ) -> TeamModel:
        """Patches an organization team.

        Args:
            organization: The name of the organization.
            team_name: The current name of the team to modify.
            new_team_name: The new name for the team.

        Returns:
            A TeamModel object representing the updated team.

        Example:
            ```python
            client.patch_organization_team(organization="My_Organization_Clan", team_name="surveyors", new_team_name="field_surveyors")
            ```
        """
        team = cast(
            TeamModel,
            self._request_json(
                "PUT",
                f"organizations/{organization}/teams/{team_name}/",
                {
                    "team": new_team_name,
                },
            ),
        )
        return team

    def delete_organization_team(self, organization: str, team_name: str) -> None:
        """Deletes an organization team.

        Args:
            organization: The name of the organization.
            team_name: The name of the team to delete.

        Example:
            ```python
            client.delete_organization_team(organization="My_Organization_Clan", team_name="field_surveyors")
            ```
        """
        self._request("DELETE", f"organizations/{organization}/teams/{team_name}/")

    def get_organization_team_members(
        self, organization: str, team_name: str
    ) -> List[TeamMemberModel]:
        """Gets a list of organization team members.

        Args:
            organization: The name of the organization.
            team_name: The name of the team.

        Returns:
            A list of TeamMemberModel objects representing the team's members.

        Example:
            ```python
            members = client.get_organization_team_members(organization="My_Organization_Clan", team_name="admins")
            for member in members:
                print(member.member)
            ```
        """
        members = cast(
            List[TeamMemberModel],
            self._request_json(
                "GET", f"organizations/{organization}/teams/{team_name}/members/"
            ),
        )
        return members

    def add_organization_team_member(
        self, organization: str, team_name: str, member_username: str
    ) -> TeamMemberModel:
        """Adds an organization team member.

        Args:
            organization: The name of the organization.
            team_name: The name of the team.
            member_username: The username of the member to add.

        Returns:
            A TeamMemberModel object representing the newly added team member.

        Example:
            ```python
            client.add_organization_team_member(organization="My_Organization_Clan", team_name="surveyors", member_username="surveyor_007")
            ```
        """
        member = cast(
            TeamMemberModel,
            self._request_json(
                "POST",
                f"organizations/{organization}/teams/{team_name}/members/",
                {"member": member_username},
            ),
        )
        return member

    def remove_organization_team_member(
        self, organization: str, team_name: str, member_username: str
    ) -> None:
        """Removes an organization team member.

        Args:
            organization: The name of the organization.
            team_name: The name of the team.
            member_username: The username of the member to remove.

        Example:
            ```python
            client.delete_organization_team_member(organization="My_Organization_Clan", team_name="surveyors", member_username="surveyor_007")
            ```
        """
        self._request(
            "DELETE",
            f"organizations/{organization}/teams/{team_name}/members/{member_username}/",
        )

    def _request_json(
        self,
        method: str,
        path: str,
        data: Any = None,
        params: Dict[str, str] = {},
        headers: Dict[str, str] = {},
        files: Optional[Dict[str, Any]] = None,
        stream: bool = False,
        skip_token: bool = False,
        allow_redirects=None,
        pagination: Pagination = Pagination(),
    ) -> Union[List, Dict]:
        result: Optional[Union[List, Dict]] = None
        is_empty_pagination = pagination.is_empty

        while True:
            resp = self._request(
                method,
                path,
                data,
                params,
                headers,
                files,
                stream,
                skip_token,
                allow_redirects,
                pagination,
            )

            payload = resp.json()

            if isinstance(payload, list):
                result = cast(List, result)

                if result:
                    result += payload
                else:
                    result = payload
            elif isinstance(payload, dict):
                result = cast(Dict, result)

                if result:
                    result = {**result, **payload}
                else:
                    result = payload
            else:
                raise NotImplementedError(
                    "Unsupported data type for paginated response."
                )

            if not is_empty_pagination:
                break

            next_url = resp.headers.get("X-Next-Page")
            if not next_url:
                break

            query_params = urlparse.parse_qs(urlparse.urlparse(next_url).query)
            pagination = Pagination(
                limit=cast(int, query_params["limit"]),
                offset=cast(int, query_params["offset"]),
            )

        return result

    def _request(
        self,
        method: str,
        path: str,
        data: Any = None,
        params: Dict[str, str] = {},
        headers: Dict[str, str] = {},
        files: Optional[Dict[str, Any]] = None,
        stream: bool = False,
        skip_token: bool = False,
        allow_redirects=None,
        pagination: Optional[Pagination] = None,
    ) -> requests.Response:
        method = method.upper()
        headers_copy = {**headers}

        assert self.url

        allow_redirects = method != "POST"

        if not skip_token and self.token:
            headers_copy["Authorization"] = f"token {self.token}"

        headers_copy["User-Agent"] = (
            f"sdk|py|{__version__} python-requests|{metadata.version('requests')}"
        )

        if not path.startswith("http"):
            if path.startswith("/"):
                path = path[1:]

            path = self.url + path

        path = add_trailing_slash_to_url(path)

        if pagination:
            limit = pagination.limit or DEFAULT_PAGINATION_LIMIT
            offset = pagination.offset or 0
            params = {
                **params,
                "limit": str(limit),
                "offset": str(offset),
            }

        request_params = {
            "method": method,
            "url": path,
            "data": data,
            "params": params,
            "headers": headers_copy,
            "files": files,
        }

        request = QfcRequest(**request_params)

        session_params = {
            "stream": stream,
            "verify": self.verify_ssl,
            # redirects from POST requests automagically turn into GET requests, so better forbid redirects
            "allow_redirects": allow_redirects,
        }

        if os.environ.get("ENVIRONMENT") == "test":
            return request.mock_response()
        else:
            response = self.session.send(request.prepare(), **session_params)

        try:
            response.raise_for_status()
        except Exception as err:
            raise QfcRequestException(response) from err

        return response



import collections
import platform
from enum import Enum
from typing import Any, Dict, List, Optional, Protocol, TypedDict

import click

from . import sdk
from .utils import format_project_table, log, print_json

QFIELDCLOUD_DEFAULT_URL = "https://app.qfield.cloud/api/v1/"


class OutputFormat(Enum):
    HUMAN = "HUMAN"
    JSON = "JSON"


class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)
        self.commands = commands or collections.OrderedDict()

    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except sdk.QfcRequestException as err:
            click.echo(str(err))
        except sdk.QfcException as err:
            click.echo(str(err))

    def list_commands(self, ctx):
        return self.commands


class ContextObject(TypedDict):
    client: sdk.Client
    format_json: bool


class Context(Protocol):
    obj: ContextObject


def paginated(command):
    command = click.option(
        "-o",
        "--offset",
        type=int,
        default=None,
        is_flag=False,
        help="Offsets the given number of records in the paginated JSON response.",
    )(command)
    command = click.option(
        "-l",
        "--limit",
        type=int,
        default=None,
        is_flag=False,
        help="Limits the number of records to return in the paginated JSON response.",
    )(command)
    return command


@click.group(cls=OrderedGroup)
@click.option(
    "-U",
    "--url",
    envvar="QFIELDCLOUD_URL",
    default=QFIELDCLOUD_DEFAULT_URL,
    type=str,
    help=f"URL to the QFieldCloud API endpoint. If not passed, gets the value from QFIELDCLOUD_URL environment variable. Default: {QFIELDCLOUD_DEFAULT_URL}",
)
@click.option(
    "-u",
    "--username",
    envvar="QFIELDCLOUD_USERNAME",
    type=str,
    help="Username or email.",
)
@click.option("-p", "--password", envvar="QFIELDCLOUD_PASSWORD", type=str)
@click.option(
    "-t", "--token", envvar="QFIELDCLOUD_TOKEN", type=str, help="Session token."
)
@click.option(
    "--json/--human",
    "format_json",
    help="Output the result as newline formatted json. Default: False",
)
@click.option(
    "--verify-ssl/--no-verify-ssl",
    "verify_ssl",
    default=True,
    envvar="QFIELDCLOUD_VERIFY_SSL",
    help="Verify SSL. Default: True",
)
@click.version_option(sdk.__version__)
@click.pass_context
def cli(
    ctx: click.Context,
    url: str,
    username: str,
    password: str,
    token: str,
    format_json: bool,
    verify_ssl: bool,
):
    """The official QFieldCloud CLI tool. Allows interaction with the QFieldCloud server API.

    Environment:

        Environment variables can be used instead of passing some common global options.

        QFIELDCLOUD_URL - QFieldCloud API endpoint URL

        QFIELDCLOUD_USERNAME - QFieldCloud username or email. Requires `QFIELDCLOUD_PASSWORD` to be set.

        QFIELDCLOUD_PASSWORD - Password. Requires `QFIELDCLOUD_USERNAME` to be set.

        QFIELDCLOUD_TOKEN - Token that can be used instead of passing username and password. It can be obtained by running `qfieldcloud-cli login`.

        QFIELDCLOUD_VERIFY_SSL - When set to `0` has the same effect as passing `--no-verify-ssl`.


    Examples:

        qfieldcloud-cli login user pass

        qfieldcloud-cli -u user -p pass -U https://app.qfield.cloud/api/v1/ list-projects
    """
    ctx.ensure_object(dict)
    ctx.obj["client"] = sdk.Client(url, verify_ssl, token=token)
    ctx.obj["format_json"] = format_json

    if username or password:
        # Guard against parent 'pre-login' for login subcommand
        if ctx.invoked_subcommand != "login":
            ctx.obj["client"].login(username, password)


@cli.command()
@click.argument("username", envvar="QFIELDCLOUD_USERNAME", required=True)
@click.argument("password", envvar="QFIELDCLOUD_PASSWORD", required=True)
@click.pass_context
def login(ctx: Context, username, password) -> None:
    """Login to QFieldCloud."""

    user_data = ctx.obj["client"].login(username, password)

    if ctx.obj["format_json"]:
        print_json(user_data)
    else:
        log(f"Log in {username}…")
        log(f"Welcome to QFieldCloud, {user_data['username']}.")
        log(
            "QFieldCloud has generated a secret token to identify you. "
            "Put the token in your in the environment using the following code, "
            "so you do not need to write your username and password again:"
        )
        if platform.system() == "Windows":
            log(f"set QFIELDCLOUD_TOKEN={user_data['token']}")
        else:
            log(f'export QFIELDCLOUD_TOKEN="{user_data["token"]}"')


@cli.command()
@click.pass_context
def logout(ctx):
    """Logout and expire the token."""

    payload = ctx.obj["client"].logout()

    if ctx.obj["format_json"]:
        print_json(payload)
    else:
        log("Log out…")
        log(payload["detail"])


@cli.command()
@click.pass_context
def status(ctx: Context):
    """Check the status of the QFieldCloud server."""
    log("Checking server status...")
    client: sdk.Client = ctx.obj["client"]
    is_json: bool = ctx.obj["format_json"]

    status_info = client.check_server_status()

    if is_json:
        print_json(status_info)
    else:
        log(click.style("Server Status:", bold=True))

        if isinstance(status_info, dict):
            for key, value in status_info.items():
                log(f"  {key.replace('_', ' ').capitalize()}: {value}")
        else:
            log(str(status_info))


@cli.command()
@paginated
@click.option(
    "--include-public/--no-public",
    default=False,
    is_flag=True,
    help="Includes the public project in the list. Default: False",
)
@click.pass_context
def list_projects(ctx: Context, include_public: bool, **opts) -> None:
    """List QFieldCloud projects."""

    projects: List[Dict[str, Any]] = ctx.obj["client"].list_projects(
        include_public,
        sdk.Pagination(**opts),
    )

    if ctx.obj["format_json"]:
        print_json(projects)
    else:
        log("Listing projects…")
        if projects:
            log("Projects the current user has access to:")
            log(format_project_table(projects))
        else:
            log("User does not have any projects yet.")


@cli.command()
@click.argument("project_id")
@click.pass_context
def get_project(ctx: Context, project_id: str) -> None:
    """Get QFieldCloud project data."""

    project: Dict[str, Any] = ctx.obj["client"].get_project(project_id)

    if ctx.obj["format_json"]:
        print_json(project)
    else:
        if project:
            log("Project data:")
            log(format_project_table([project]))
        else:
            log("User does not have access to projects yet.")


@cli.command()
@click.argument("project_id")
@click.pass_context
def get_project_seed(ctx: Context, project_id: str) -> None:
    """Get QFieldCloud project seed data."""

    project_seed: Dict[str, Any] = ctx.obj["client"].get_project_seed(project_id)

    if ctx.obj["format_json"]:
        print_json(project_seed)
    else:
        if project_seed:
            log("Project name: {}".format(project_seed["name"]))
            log("Project CRS: {}".format(project_seed["crs"]))
            log(
                "Project extent: {}".format(
                    ", ".join(map(lambda n: str(n), project_seed["extent"]))
                )
            )
            log(
                "Project basemaps: {}".format(len(project_seed["settings"]["basemaps"]))
            )
            log("Project XLSForm: {}".format(bool(project_seed["settings"]["xlsform"])))
        else:
            log("User does not have access to projects yet.")


@cli.command()
@click.argument("project_id")
@click.argument("destination_dir")
@click.pass_context
def get_project_seed_xlsform(
    ctx: Context,
    project_id: str,
    destination_dir: str,
) -> None:
    """Get QFieldCloud project seed XLSForm file."""

    xlsform_filename = ctx.obj["client"].get_project_seed_xlsform(
        project_id, destination_dir
    )

    if ctx.obj["format_json"]:
        print_json(xlsform_filename)
    else:
        if xlsform_filename:
            log(f"XLSForm seed file downloaded to: {xlsform_filename}")
        else:
            log("No XLSForm seed file found for the project.")


@cli.command()
@click.argument("project_id")
@click.option(
    "--skip-metadata/--no-skip-metadata",
    "skip_metadata",
    default=True,
    help="Skip requesting for additional metadata (currently the `sha256` checksum) for each version. Default: --skip-metadata",
)
@click.pass_context
def list_files(ctx: Context, project_id, skip_metadata):
    """List QFieldCloud project files."""

    files = ctx.obj["client"].list_remote_files(project_id, skip_metadata)

    if ctx.obj["format_json"]:
        print_json(files)
    else:
        log(f'Getting file list for "{project_id}"…')
        if files:
            log(f'Files for project "{project_id}":')
            for file in files:
                log(f"{file['last_modified']}\t{file['name']}")
        else:
            log(f'No files within project "{project_id}"')


@cli.command()
@click.argument("name")
@click.option(
    "--owner",
    "owner",
    help="Owner of the project. If omitted, the current user is the owner.",
)
@click.option("--description", "description", help="Description of the project.")
@click.option(
    "--is-public/--is-private", "is_public", help="Mark the project as public."
)
@click.pass_context
def create_project(ctx: Context, name, owner, description, is_public):
    """Creates a new empty QFieldCloud project."""

    project = ctx.obj["client"].create_project(
        name, owner, description=description, is_public=is_public
    )

    if ctx.obj["format_json"]:
        print_json(project)
    else:
        log("Creating project {}…".format(f"{owner}/{name}" if owner else name))
        log("Created project:")
        log(format_project_table([project]))


@cli.command()
@click.argument("project_id")
@click.pass_context
def delete_project(ctx: Context, project_id):
    """Deletes a QFieldCloud project."""

    payload = ctx.obj["client"].delete_project(project_id)

    if ctx.obj["format_json"]:
        # print_json(payload)
        print(payload, payload.content)
    else:
        log(f'Deleting project "{project_id}"…')
        log(f'Deleted project "{project_id}".')


@cli.command()
@click.argument("project_id")
@click.argument("project_path")
@click.option(
    "--filter",
    "filter_glob",
    help="Do not upload the whole project, but only the files which match the glob.",
)
@click.option(
    "--throw-on-error/--no-throw-on-error",
    help="If any project file upload fails stop uploading the rest. Default: False",
)
@click.option(
    "--force/--no-force",
    help="Upload file even if it already exists remotely. Default: False",
    default=False,
)
@click.pass_context
def upload_files(
    ctx: Context,
    project_id: str,
    project_path: str,
    filter_glob: str,
    throw_on_error: bool,
    force: bool,
) -> None:
    """Upload files to a QFieldCloud project."""

    files = ctx.obj["client"].upload_files(
        project_id,
        sdk.FileTransferType.PROJECT,
        project_path,
        filter_glob=filter_glob,
        throw_on_error=throw_on_error,
        show_progress=True,
        force=force,
    )

    if ctx.obj["format_json"]:
        print_json(files)
    else:
        log(f'Uploading files "{project_id}" from {project_path}…')
        if files:
            log(f"Upload finished after uploading {len(files)}.")
            for file in files:
                if file.get("error"):
                    log(f'File "{file["name"]}" failed to upload: {file["error"]} .')
        else:
            log("Nothing to upload.")


@cli.command()
@click.argument("project_id")
@click.argument("local_dir")
@click.option(
    "--filter",
    "filter_glob",
    help="Do not download the whole project, but only the files which match the glob.",
)
@click.option(
    "--throw-on-error/--no-throw-on-error",
    help="If any project file downloads fails stop downloading the rest. Default: False",
)
@click.option(
    "--force-download/--no-force-download",
    help="Download file even if it already exists locally. Default: False",
)
@click.pass_context
def download_files(
    ctx: Context, project_id, local_dir, filter_glob, throw_on_error, force_download
):
    """Download QFieldCloud project files."""

    files = ctx.obj["client"].download_project(
        project_id,
        local_dir,
        filter_glob,
        throw_on_error,
        show_progress=True,
        force_download=force_download,
    )

    if ctx.obj["format_json"]:
        print_json(files)
    else:
        log(f'Downloading project "{project_id}" files to {local_dir}…')
        if files:
            count = 0
            for file in files:
                if file.get("error"):
                    log(f'File "{file["name"]}" failed to download: {file["error"]} .')
                else:
                    count += 1

            log(f"Downloaded {count} file(s).")
        else:
            if filter_glob:
                log(f"No files to download for project {project_id} at {filter_glob}")
            else:
                log(f"No files to download for project {project_id}")


@cli.command()
@click.argument("project_id")
@click.option(
    "--name",
    help="New project name",
)
@click.option(
    "--description",
    help="New project description",
)
@click.option(
    "--owner",
    help="Transfer the project to a new owner",
)
@click.option(
    "--is-public/--is-no-public",
    is_flag=True,
    help="Whether the project shall be public",
)
@click.pass_context
def patch_project(
    ctx: Context,
    project_id: str,
    name: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    is_public: Optional[bool] = None,
) -> None:
    """Patch the project with new data. Pass only the parameters that shall be changed."""

    project = ctx.obj["client"].patch_project(
        project_id, name=name, owner=owner, description=description, is_public=is_public
    )

    if ctx.obj["format_json"]:
        print_json(project)
    else:
        log("Patched project:")
        log(format_project_table([project]))


@cli.command()
@click.argument("project_id")
@click.argument("paths", nargs=-1, required=True)
@click.option(
    "--throw-on-error/--no-throw-on-error",
    help="If any project file delete operations fails stop, stop deleting the rest. Default: False",
)
@click.pass_context
def delete_files(ctx: Context, project_id, paths, throw_on_error):
    """Delete QFieldCloud project files."""

    paths_result = ctx.obj["client"].delete_files(project_id, paths, throw_on_error)

    if ctx.obj["format_json"]:
        print_json(paths_result)
    else:
        log(f'Deleting project "{project_id}" files…')


@cli.command()
@click.argument("project_id")
@click.argument("filename")
@click.option(
    "--version-id",
    help="Pass to delete a specific file version. Otherwise the file and all it's versions will be deleted.",
)
@click.pass_context
def delete_file(
    ctx: Context, project_id: str, filename: str, version_id: Optional[str]
) -> None:
    """Delete QFieldCloud project files."""

    log(f'Deleting file "{filename}" in project "{project_id}"…')

    _resp = ctx.obj["client"].delete_file(project_id, filename, version_id)

    log(f'Successfully deleted file "{filename}" in project "{project_id}"!')


@cli.command()
@click.argument("project_id")
@click.option(
    "--type",
    "job_type",
    type=sdk.JobTypes,
    help="Job type. One of package, delta_apply or process_projectfile.",
)
@paginated
@click.pass_context
def list_jobs(ctx: Context, project_id, job_type: Optional[sdk.JobTypes], **opts):
    """List project jobs."""

    jobs: List[Dict] = ctx.obj["client"].list_jobs(
        project_id,
        job_type,
        sdk.Pagination(**opts),
    )

    if ctx.obj["format_json"]:
        print_json(jobs)
    else:
        log(f'Listing project "{project_id}" jobs…')
        for job in jobs:
            log(
                f'Job "{job["id"]}" of project "{project_id}" is of type "{job["type"]}" and has status "{job["status"]}".'
            )


@cli.command()
@click.argument("project_id")
@click.argument("job_type", type=sdk.JobTypes)
@click.option(
    "--force/--no-force",
    default=False,
    help="Should force creating a new job. Default: False",
)
@click.pass_context
def job_trigger(ctx: Context, project_id, job_type, force):
    """Triggers a new job."""

    status = ctx.obj["client"].job_trigger(project_id, job_type, force)

    if ctx.obj["format_json"]:
        print_json(status)
    else:
        log(f'Triggering "{job_type}" job for project "{project_id}"…')
        log(
            f'Job of type "{job_type}" triggered for project "{project_id}": {status["id"]}'
        )


@cli.command()
@click.argument("job_id")
@click.pass_context
def job_status(ctx: Context, job_id):
    """Get job status."""

    status = ctx.obj["client"].job_status(job_id)

    if ctx.obj["format_json"]:
        print_json(status)
    else:
        log(f'Getting job "{job_id}" status…')
        log(f"Job status for {job_id}: {status['status']}")


@cli.command(short_help="Push a delta file to a project.")
@click.argument("project_id")
@click.argument("delta_filename", type=click.Path(exists=True))
@click.pass_context
def delta_push(ctx: Context, project_id: str, delta_filename: str) -> None:
    """Push a delta file to a project with PROJECT_ID."""

    response = ctx.obj["client"].push_delta(project_id, delta_filename)

    if ctx.obj["format_json"]:
        print_json(response)
    else:
        log(f'Pushing delta file "{delta_filename}" to project "{project_id}"…')
        log(f'Delta file "{delta_filename}" pushed to project "{project_id}".')


@cli.command()
@click.argument("project_id")
@click.pass_context
def package_latest(ctx: Context, project_id):
    """Check project packaging status."""

    status = ctx.obj["client"].package_latest(project_id)

    if ctx.obj["format_json"]:
        print_json(status)
    else:
        log(f'Getting the latest project "{project_id}" package info…')
        log(f"Packaging status for {project_id}: {status['status']}")
        if status["layers"] is None:
            if status["status"] == "failed":
                log("Packaging have never been triggered on this project. Please run:")
                log(f"qfieldcloud-cli job-trigger {project_id} package")
            return

        for layer_obj in status["layers"].values():
            if layer_obj["is_valid"]:
                log(
                    f'Layer "{layer_obj["name"]}" is valid, finished with status: {layer_obj["error_code"]}'
                )
            else:
                log(
                    f'Invalid layer "{layer_obj["name"]}", status: {layer_obj["error_code"]}'
                )


@cli.command()
@click.argument("project_id")
@click.argument("local_dir")
@click.option(
    "--filter",
    "filter_glob",
    help="Do not download the whole packaged project, but only the files which match the glob.",
)
@click.option(
    "--throw-on-error/--no-throw-on-error",
    help="If any packaged file downloads fails stop downloading the rest. Default: False",
)
@click.option(
    "--force-download/--no-force-download",
    help="Download file even if it already exists locally. Default: False",
)
@click.pass_context
def package_download(
    ctx: Context, project_id, local_dir, filter_glob, throw_on_error, force_download
):
    """Download packaged QFieldCloud project files."""

    files = ctx.obj["client"].package_download(
        project_id,
        local_dir,
        filter_glob,
        throw_on_error,
        show_progress=True,
        force_download=force_download,
    )

    if ctx.obj["format_json"]:
        print_json(files)
    else:
        log(
            f'Downloading the latest project "{project_id}" package files to {local_dir}…'
        )
        if files:
            log(f"Download status of packaged files in project {project_id}:")
            for file in files:
                log(f"{file['status'].value}\t{file['name']}")
        else:
            if filter_glob:
                log(
                    f"No packaged files to download for project {project_id} at {filter_glob}"
                )
            else:
                log(f"No packaged files to download for project {project_id}")


@cli.command(short_help="Get a list of project collaborators.")
@click.argument("project_id")
@click.pass_context
def collaborators_get(ctx: Context, project_id: str) -> None:
    """Get a list of project collaborators for specific project with PROJECT_ID."""
    collaborators = ctx.obj["client"].get_project_collaborators(project_id)

    if ctx.obj["format_json"]:
        print_json(collaborators)
    else:
        log(f'Collaborators for project with id "{project_id}":')
        for collaborator in collaborators:
            log(f"{collaborator['collaborator']}\t{collaborator['role']}")


@cli.command(short_help="Add a project collaborator.")
@click.argument("project_id")
@click.argument("username")
@click.argument("role", type=sdk.ProjectCollaboratorRole)
@click.pass_context
def collaborators_add(
    ctx: Context, project_id: str, username: str, role: sdk.ProjectCollaboratorRole
) -> None:
    """Add collaborator with USERNAME with specific ROLE to a project with PROJECT_ID. Possible ROLE values: admin, manager, editor, reporter, reader."""
    collaborator = ctx.obj["client"].add_project_collaborator(
        project_id, username, role
    )

    if ctx.obj["format_json"]:
        print_json(collaborator)
    else:
        log(
            f'Collaborator "{collaborator["collaborator"]}" added to project with id "{collaborator["project_id"]}" with role "{collaborator["role"]}".'
        )


@cli.command(short_help="Remove a project collaborator.")
@click.argument("project_id")
@click.argument("username")
@click.pass_context
def collaborators_remove(ctx: Context, project_id: str, username: str) -> None:
    """Remove collaborator with USERNAME from project with PROJECT_ID."""
    ctx.obj["client"].remove_project_collaborator(project_id, username)

    if not ctx.obj["format_json"]:
        log(f'Collaborator "{username}" removed project with id "{project_id}".')


@cli.command(short_help="Change project collaborator role.")
@click.argument("project_id")
@click.argument("username")
@click.argument("role", type=sdk.ProjectCollaboratorRole)
@click.pass_context
def collaborators_patch(
    ctx: Context, project_id: str, username: str, role: sdk.ProjectCollaboratorRole
) -> None:
    """Change collaborator with USERNAME to new ROLE in project with PROJECT_ID. Possible ROLE values: admin, manager, editor, reporter, reader."""
    collaborator = ctx.obj["client"].patch_project_collaborators(
        project_id, username, role
    )

    if ctx.obj["format_json"]:
        print_json(collaborator)
    else:
        log(
            f'Collaborator "{collaborator["collaborator"]}" added to project with id "{collaborator["project_id"]}" with role "{collaborator["role"]}".'
        )


@cli.command(short_help="Get a list organization members.")
@click.argument("organization")
@click.pass_context
def members_get(ctx: Context, organization: str) -> None:
    """Get a list of ORGANIZATION members."""
    memberships = ctx.obj["client"].get_organization_members(organization)

    if ctx.obj["format_json"]:
        print_json(memberships)
    else:
        log(f'Members of organization "{organization}":')
        for membership in memberships:
            log(f"{membership['member']}\t{membership['role']}")


@cli.command(short_help="Add an organization member.")
@click.argument("organization")
@click.argument("username")
@click.argument("role", type=sdk.OrganizationMemberRole)
@click.option("--public/--no-public", "is_public")
@click.pass_context
def members_add(
    ctx: Context,
    organization: str,
    username: str,
    role: sdk.OrganizationMemberRole,
    is_public: bool,
) -> None:
    """Add member with USERNAME with ROLE to ORGANIZATION. Possible ROLE values: admin, member."""
    membership = ctx.obj["client"].add_organization_member(
        organization, username, role, is_public
    )

    if ctx.obj["format_json"]:
        print_json(membership)
    else:
        log(
            f'Member "{membership["member"]}" added to organization "{membership["organization"]}" with role "{membership["role"]}".'
        )


@cli.command(short_help="Remove an organization member.")
@click.argument("organization")
@click.argument("username")
@click.pass_context
def members_remove(ctx: Context, organization: str, username: str) -> None:
    """Remove member with USERNAME from ORGANIZATION."""
    ctx.obj["client"].remove_organization_members(organization, username)

    if not ctx.obj["format_json"]:
        log(f'Member "{username}" removed organization "{organization}".')


@cli.command(short_help="Change organization member role.")
@click.argument("organization")
@click.argument("username")
@click.argument("role", type=sdk.OrganizationMemberRole)
@click.pass_context
def members_patch(
    ctx: Context, organization: str, username: str, role: sdk.OrganizationMemberRole
) -> None:
    """Change member with USERNAME to new ROLE in ORGANIZATION. Possible ROLE values: admin, member."""
    membership = ctx.obj["client"].patch_organization_members(
        organization, username, role
    )

    if ctx.obj["format_json"]:
        print_json(membership)
    else:
        log(
            f'Member "{membership["member"]}" changed role in organization "{membership["organization"]}" to role "{membership["role"]}".'
        )


@cli.command(short_help="Get a list of organization teams.")
@click.argument("organization")
@click.pass_context
def teams_list(ctx: Context, organization: str) -> None:
    """Get a list of organization teams."""
    teams_list = ctx.obj["client"].get_organization_teams(organization)

    if ctx.obj["format_json"]:
        print_json(teams_list)
    else:
        log(f'Teams members in organization "{organization}":')
        for object_team in teams_list:
            log(f"{object_team['team']}")


@cli.command(name="teams-create", short_help="Create an organization team.")
@click.argument("organization")
@click.argument("team_name")
@click.pass_context
def teams_create(ctx: Context, organization: str, team_name: str) -> None:
    """Create a new team named TEAM_NAME in ORGANIZATION."""
    object_team = ctx.obj["client"].create_organization_team(organization, team_name)

    if ctx.obj["format_json"]:
        print_json(object_team)
    else:
        log(f'Team "{object_team["team"]}" created in organization "{organization}".')


@cli.command(name="teams-get", short_help="Get a list of teams on an organization.")
@click.argument("organization")
@click.argument("team_name")
@click.pass_context
def teams_get(ctx: Context, organization: str, team_name: str) -> None:
    """Get details of team TEAM_NAME in ORGANIZATION."""
    object_team = ctx.obj["client"].get_organization_team(organization, team_name)

    if ctx.obj["format_json"]:
        print_json(object_team)
    else:
        log(
            f'Team "{object_team["team"]}" in organization "{object_team["organization"]}":'
        )
        log(f"  Members: {', '.join(object_team['members'])}")


@cli.command(name="teams-patch", short_help="Rename an organization team.")
@click.argument("organization")
@click.argument("team_name")
@click.option("--name", "new_team_name")
@click.pass_context
def teams_patch(
    ctx: Context, organization: str, team_name: str, new_team_name: str
) -> None:
    """Rename team TEAM_NAME to NEW_TEAM_NAME in ORGANIZATION."""
    object_team = ctx.obj["client"].patch_organization_team(
        organization, team_name, new_team_name
    )

    if ctx.obj["format_json"]:
        print_json(object_team)
    else:
        log(
            f'Team "{team_name}" in organization "{organization}" was renamed to "{object_team["team"]}".'
        )


@cli.command(name="teams-delete", short_help="Delete an organization team.")
@click.argument("organization")
@click.argument("team_name")
@click.pass_context
def teams_delete(ctx: Context, organization: str, team_name: str) -> None:
    """Delete team TEAM_NAME from ORGANIZATION."""
    ctx.obj["client"].delete_organization_team(organization, team_name)

    if not ctx.obj["format_json"]:
        log(f'Team "{team_name}" was deleted from organization "{organization}".')


@cli.command(
    name="team-members-list", short_help="List members of an organization team."
)
@click.argument("organization")
@click.argument("team_name")
@click.pass_context
def team_members_list(ctx: Context, organization: str, team_name: str) -> None:
    """List members of team TEAM_NAME in ORGANIZATION"""
    members = ctx.obj["client"].get_organization_team_members(organization, team_name)
    if ctx.obj["format_json"]:
        print_json(members)
    else:
        log(f'Members of team "{team_name}" in organization "{organization}":')

        for object_member in members:
            log(object_member["member"])


@cli.command(
    name="team-members-add", short_help="Add a member to an organization team."
)
@click.argument("organization")
@click.argument("team_name")
@click.argument("member_username")
@click.pass_context
def team_members_add(
    ctx: Context, organization: str, team_name: str, member_username: str
) -> None:
    """Add member MEMBER_USERNAME to team TEAM_NAME in ORGANIZATION."""
    object_member = ctx.obj["client"].add_organization_team_member(
        organization, team_name, member_username
    )

    if ctx.obj["format_json"]:
        print_json(object_member)
    else:
        log(
            f'Member "{object_member["member"]}" added to team "{team_name}" in organization "{organization}".'
        )


@cli.command(
    name="team-members-remove", short_help="Remove a member from an organization team."
)
@click.argument("organization")
@click.argument("team_name")
@click.argument("member_username")
@click.pass_context
def team_members_remove(
    ctx: Context, organization: str, team_name: str, member_username: str
) -> None:
    """Remove member MEMBER_USERNAME from team TEAM_NAME in ORGANIZATION."""
    ctx.obj["client"].remove_organization_team_member(
        organization, team_name, member_username
    )

    if not ctx.obj["format_json"]:
        log(
            f'Member "{member_username}" removed from team "{team_name}" in organization "{organization}".'
        )



import unittest

from click.testing import CliRunner

from qfieldcloud_sdk.cli import QFIELDCLOUD_DEFAULT_URL, cli
from qfieldcloud_sdk.sdk import Client, Pagination


class TestSDK(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = Client(QFIELDCLOUD_DEFAULT_URL)

    def test_paginated_list_projects(self):
        results = self.client.list_projects(limit=20)
        self.assertTrue(0 < len(results) and len(results) <= 20)

    def test_paginated_list_projects_include_public(self):
        results = self.client.list_projects(
            include_public=True, pagination=Pagination(limit=200)
        )
        self.assertTrue(0 < len(results) and len(results) <= 50)


class TestCLI(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.runner = CliRunner()

    def test_list_projects(self):
        result = self.runner.invoke(
            cli,
            [
                "list-projects",
                "--include-public",
                "--offset",
                5,
                "--limit",
                100,
            ],
            catch_exceptions=False,
        )
        self.assertEqual(result.exit_code, 0)

    def test_list_jobs(self):
        result = self.runner.invoke(
            cli,
            ["list-jobs", "my_project_id", "--limit", 10, "--offset", 5],
            catch_exceptions=False,
        )
        self.assertEqual(result.exit_code, 0)



import json
from typing import Any

import requests


class QfcMockItem(dict):
    def __getitem__(self, k: str) -> Any:
        if k == "id":
            return super().__getitem__("id")
        else:
            return k


class QfcMockResponse(requests.Response):
    def __init__(self, **kwargs):
        self.request_kwargs = kwargs
        self.url = kwargs["url"]
        self.limit = kwargs.get("limit", 5)
        self.total = self.limit * 2
        self.headers = {
            "X-Total-Count": self.total,
        }

        limit = kwargs["params"].get("limit")
        offset = kwargs["params"].get("offset", 0)
        prev_url = None
        next_url = None
        if limit:
            if offset == 0:
                prev_url = None
                next_url = f"{self.url}?limit={limit}&offset={limit}"
            else:
                prev_url = f"{self.url}?limit={limit}&offset=0"
                next_url = None

        self.headers["X-Previous-Page"] = prev_url
        self.headers["X-Next-Page"] = next_url

    def json(self):
        if self.request_kwargs["method"] == "GET":
            return [QfcMockItem(id=n) for n in range(self.total)]
        else:
            return QfcMockItem(id="test_id", **self.request_kwargs)


class QfcRequest(requests.Request):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kwargs = kwargs

    def mock_response(self) -> QfcMockResponse:
        return QfcMockResponse(**self.kwargs)


class QfcException(Exception):
    def __init__(self, reason: str, *args):
        super().__init__(reason, *args)


class QfcRequestException(QfcException):
    def __init__(self, response: requests.Response, *args):
        super().__init__(str(response), *args)
        self.response = response

        try:
            json_content = response.json()
            json_content = json.dumps(json_content, sort_keys=True, indent=2)
        except Exception:
            json_content = ""

        self.reason = f'Requested "{response.url}" and got "{response.status_code} {response.reason}":\n{json_content or response.content.decode()}'

    def __str__(self) -> str:
        return self.reason

    def __repr__(self) -> str:
        return self.reason



import hashlib
import json
import os
import sys
from typing import List
from urllib.parse import urlparse, urlunparse


def print_json(data):
    """Pretty print as JSON string"""
    print(json.dumps(data, sort_keys=True, indent=2))


def log(*msgs):
    """Print text chunks to stderr"""
    print(*msgs, file=sys.stderr)


def get_md5sum(filename: str) -> str:
    """Calculate md5sum of a file.

    Currently unused but will be revived in the upcoming versions.
    """
    BLOCKSIZE = 65536
    hasher = hashlib.md5()
    with open(filename, "rb") as f:
        buf = f.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = f.read(BLOCKSIZE)
    return hasher.hexdigest()


def calc_etag(filename: str, part_size: int = 8 * 1024 * 1024) -> str:
    """Calculate ETag as in Object Storage (S3) of a local file.

    ETag is a MD5. But for the multipart uploaded files, the MD5 is computed from the concatenation of the MD5s of each uploaded part.

    See the inspiration of this implementation here: https://stackoverflow.com/a/58239738/1226137

    Args:
        filename (str): the local filename
        part_size (int): the size of the Object Storage part. Most Object Storages use 8MB. Defaults to 8*1024*1024.

    Returns:
        str: the calculated ETag value
    """
    with open(filename, "rb") as f:
        file_size = os.fstat(f.fileno()).st_size

        if file_size <= part_size:
            BLOCKSIZE = 65536
            hasher = hashlib.md5()

            buf = f.read(BLOCKSIZE)
            while len(buf) > 0:
                hasher.update(buf)
                buf = f.read(BLOCKSIZE)

            return hasher.hexdigest()
        else:
            # Say you uploaded a 14MB file and your part size is 5MB.
            # Calculate 3 MD5 checksums corresponding to each part, i.e. the checksum of the first 5MB, the second 5MB, and the last 4MB.
            # Then take the checksum of their concatenation.
            # Since MD5 checksums are hex representations of binary data, just make sure you take the MD5 of the decoded binary concatenation, not of the ASCII or UTF-8 encoded concatenation.
            # When that's done, add a hyphen and the number of parts to get the ETag.
            md5sums = []
            for data in iter(lambda: f.read(part_size), b""):
                md5sums.append(hashlib.md5(data).digest())

            final_md5sum = hashlib.md5(b"".join(md5sums))

            return "{}-{}".format(final_md5sum.hexdigest(), len(md5sums))


def format_table(headers: List[str], data: List[List]) -> str:
    length_by_column: List[int] = []

    for col in headers:
        length_by_column.append(len(col))

    for row in data:
        for idx, col in enumerate(row):
            length_by_column[idx] = max(length_by_column[idx], len(str(col)))

    row_tmpl = "|"
    for col_length in length_by_column:
        row_tmpl += " {:<" + str(col_length) + "} |"

    result = row_tmpl.format(*headers)
    result += "\r\n"
    result += "-" * (sum(length_by_column) + len(headers) * 3 + 1)

    for row in data:
        result += "\r\n"
        result += row_tmpl.format(*row)

    return result


def format_project_table(projects: List) -> str:
    data = []

    for project in projects:
        data.append(
            [
                project["id"],
                project["owner"] + "/" + project["name"],
                project["is_public"],
                project["description"],
            ]
        )

    return format_table(
        headers=["ID", "OWNER/NAME", "IS PUBLIC", "DESCRIPTION"],
        data=data,
    )


def add_trailing_slash_to_url(url: str) -> str:
    """
    Add a trailing slash to a URL if it doesn't already have one.

    Args:
        url (str): The URL to process

    Returns:
        str: URL with trailing slash added

    Examples:
        >>> add_trailing_slash("https://example.com")
        'https://example.com/'
        >>> add_trailing_slash("https://example.com/")
        'https://example.com/'
        >>> add_trailing_slash("https://example.com/path")
        'https://example.com/path/'
        >>> add_trailing_slash("https://example.com/file.txt")
        'https://example.com/file.txt/'
    """
    if not url:
        return url

    parsed = urlparse(url)

    # Add trailing slash to path if it doesn't end with one
    path = parsed.path
    if not path.endswith("/"):
        path += "/"

    # Reconstruct the URL with the modified path
    modified_parsed = parsed._replace(path=path)
    return urlunparse(modified_parsed)
